#pragma once

#include <unibilium.h>  // IWYU pragma: keep

#include "nvim/api/private/defs.h"  // IWYU pragma: keep

#ifdef INCLUDE_GENERATED_DECLARATIONS
# include "tui/terminfo.h.generated.h"
#endif
