#pragma once

#include <stddef.h>  // IWYU pragma: keep
#include <time.h>  // IWYU pragma: keep

#include "nvim/os/time_defs.h"  // IWYU pragma: keep

#ifdef INCLUDE_GENERATED_DECLARATIONS
# include "os/time.h.generated.h"
#endif
