#pragma once

#include <stddef.h>  // IWYU pragma: keep

#ifdef MSWIN
# include "nvim/api/private/defs.h"  // IWYU pragma: keep
#endif

#ifdef INCLUDE_GENERATED_DECLARATIONS
# include "os/process.h.generated.h"
#endif
