#pragma once

typedef enum {
  KITTY_KEY_ESCAPE               = 57344,
  KITTY_KEY_ENTER                = 57345,
  KITTY_KEY_TAB                  = 57346,
  KITTY_KEY_BACKSPACE            = 57347,
  KITTY_KEY_INSERT               = 57348,
  KITTY_KEY_DELETE               = 57349,
  KITTY_KEY_LEFT                 = 57350,
  KITTY_KEY_RIGHT                = 57351,
  KITTY_KEY_UP                   = 57352,
  KITTY_KEY_DOWN                 = 57353,
  KITTY_KEY_PAGE_UP              = 57354,
  KITTY_KEY_PAGE_DOWN            = 57355,
  KITTY_KEY_HOME                 = 57356,
  KITTY_KEY_END                  = 57357,
  KITTY_KEY_CAPS_LOCK            = 57358,
  KITTY_KEY_SCROLL_LOCK          = 57359,
  KITTY_KEY_NUM_LOCK             = 57360,
  KITTY_KEY_PRINT_SCREEN         = 57361,
  KITTY_KEY_PAUSE                = 57362,
  KITTY_KEY_MENU                 = 57363,
  KITTY_KEY_F1                   = 57364,
  KITTY_KEY_F2                   = 57365,
  KITTY_KEY_F3                   = 57366,
  KITTY_KEY_F4                   = 57367,
  KITTY_KEY_F5                   = 57368,
  KITTY_KEY_F6                   = 57369,
  KITTY_KEY_F7                   = 57370,
  KITTY_KEY_F8                   = 57371,
  KITTY_KEY_F9                   = 57372,
  KITTY_KEY_F10                  = 57373,
  KITTY_KEY_F11                  = 57374,
  KITTY_KEY_F12                  = 57375,
  KITTY_KEY_F13                  = 57376,
  KITTY_KEY_F14                  = 57377,
  KITTY_KEY_F15                  = 57378,
  KITTY_KEY_F16                  = 57379,
  KITTY_KEY_F17                  = 57380,
  KITTY_KEY_F18                  = 57381,
  KITTY_KEY_F19                  = 57382,
  KITTY_KEY_F20                  = 57383,
  KITTY_KEY_F21                  = 57384,
  KITTY_KEY_F22                  = 57385,
  KITTY_KEY_F23                  = 57386,
  KITTY_KEY_F24                  = 57387,
  KITTY_KEY_F25                  = 57388,
  KITTY_KEY_F26                  = 57389,
  KITTY_KEY_F27                  = 57390,
  KITTY_KEY_F28                  = 57391,
  KITTY_KEY_F29                  = 57392,
  KITTY_KEY_F30                  = 57393,
  KITTY_KEY_F31                  = 57394,
  KITTY_KEY_F32                  = 57395,
  KITTY_KEY_F33                  = 57396,
  KITTY_KEY_F34                  = 57397,
  KITTY_KEY_F35                  = 57398,
  KITTY_KEY_KP_0                 = 57399,
  KITTY_KEY_KP_1                 = 57400,
  KITTY_KEY_KP_2                 = 57401,
  KITTY_KEY_KP_3                 = 57402,
  KITTY_KEY_KP_4                 = 57403,
  KITTY_KEY_KP_5                 = 57404,
  KITTY_KEY_KP_6                 = 57405,
  KITTY_KEY_KP_7                 = 57406,
  KITTY_KEY_KP_8                 = 57407,
  KITTY_KEY_KP_9                 = 57408,
  KITTY_KEY_KP_DECIMAL           = 57409,
  KITTY_KEY_KP_DIVIDE            = 57410,
  KITTY_KEY_KP_MULTIPLY          = 57411,
  KITTY_KEY_KP_SUBTRACT          = 57412,
  KITTY_KEY_KP_ADD               = 57413,
  KITTY_KEY_KP_ENTER             = 57414,
  KITTY_KEY_KP_EQUAL             = 57415,
  KITTY_KEY_KP_SEPARATOR         = 57416,
  KITTY_KEY_KP_LEFT              = 57417,
  KITTY_KEY_KP_RIGHT             = 57418,
  KITTY_KEY_KP_UP                = 57419,
  KITTY_KEY_KP_DOWN              = 57420,
  KITTY_KEY_KP_PAGE_UP           = 57421,
  KITTY_KEY_KP_PAGE_DOWN         = 57422,
  KITTY_KEY_KP_HOME              = 57423,
  KITTY_KEY_KP_END               = 57424,
  KITTY_KEY_KP_INSERT            = 57425,
  KITTY_KEY_KP_DELETE            = 57426,
  KITTY_KEY_KP_BEGIN             = 57427,
  KITTY_KEY_MEDIA_PLAY           = 57428,
  KITTY_KEY_MEDIA_PAUSE          = 57429,
  KITTY_KEY_MEDIA_PLAY_PAUSE     = 57430,
  KITTY_KEY_MEDIA_REVERSE        = 57431,
  KITTY_KEY_MEDIA_STOP           = 57432,
  KITTY_KEY_MEDIA_FAST_FORWARD   = 57433,
  KITTY_KEY_MEDIA_REWIND         = 57434,
  KITTY_KEY_MEDIA_TRACK_NEXT     = 57435,
  KITTY_KEY_MEDIA_TRACK_PREVIOUS = 57436,
  KITTY_KEY_MEDIA_RECORD         = 57437,
  KITTY_KEY_LOWER_VOLUME         = 57438,
  KITTY_KEY_RAISE_VOLUME         = 57439,
  KITTY_KEY_MUTE_VOLUME          = 57440,
  KITTY_KEY_LEFT_SHIFT           = 57441,
  KITTY_KEY_LEFT_CONTROL         = 57442,
  KITTY_KEY_LEFT_ALT             = 57443,
  KITTY_KEY_LEFT_SUPER           = 57444,
  KITTY_KEY_LEFT_HYPER           = 57445,
  KITTY_KEY_LEFT_META            = 57446,
  KITTY_KEY_RIGHT_SHIFT          = 57447,
  KITTY_KEY_RIGHT_CONTROL        = 57448,
  KITTY_KEY_RIGHT_ALT            = 57449,
  KITTY_KEY_RIGHT_SUPER          = 57450,
  KITTY_KEY_RIGHT_HYPER          = 57451,
  KITTY_KEY_RIGHT_META           = 57452,
  KITTY_KEY_ISO_LEVEL3_SHIFT     = 57453,
  KITTY_KEY_ISO_LEVEL5_SHIFT     = 57454,
} KittyKey;
